
/****************************************************************************/
/*                                                                          */
/*  Functions to scan (parse), print (display in human readable format),    */
/*  set (modify in dangerous ways), check integrity and compare various     */
/*  data types. At some stage this should be improved to have proper IO     */
/*  functions for each unit - instead of the half-baked scan and print.     */
/*  Problem is that types which use symbolic names are not chroot()         */
/*  safe - they require support files in jail.                              */
/*                                                                          */
/*  Warning - many of these functions are incomplete and just borrow the    */
/*  integer routines. So before porting to any arch where time_t != int     */
/*  the type handlers will have to implemented.                             */
/*                                                                          */
/****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <idsa_internal.h>

/* string handler ******************************************************* */

static int _idsa_string_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;

  result = strncmp(a->u_ptr, b->u_ptr, idsa_type_size(a->u_type));

  if (result < 0) {
    result = IDSA_COMPARE_LESS;
  } else if (result > 0) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

#ifdef TRACE
  fprintf(stderr, "compare_string(): comparing of <%s> against <%s>\n",
	  a->u_ptr, b->u_ptr);
#endif

  return result;
}

static int _idsa_string_check(IDSA_UNIT * u)
{
  u->u_ptr[idsa_type_size(u->u_type) - 1] = '\0';
  return 0;
}

static int _idsa_string_scan(IDSA_UNIT * u, char *s)
{
  int i, z;
  int result = 0;

  i = strlen(s);
  z = idsa_type_size(u->u_type);

  if (i > (z - 1)) {
    i = z - 1;
    result = 1;			/* string too long, but we try our best */
  }
  strncpy(u->u_ptr, s, i);
  u->u_ptr[i] = '\0';

  return result;
}

static int _idsa_string_print(IDSA_UNIT * u, char *s, int l, int m)
{
  int i;

  u->u_ptr[idsa_type_size(u->u_type) - 1] = '\0';
  i = strlen(u->u_ptr);
  if (l < i) {
    memcpy(s, u->u_ptr, l);	/* feeble best effort */
    return -1;
  } else {
    memcpy(s, u->u_ptr, i);
    return i;
  }
}

static int _idsa_string_set(IDSA_UNIT * u, void *p)
{
  return _idsa_string_scan(u, p);
}

static int _idsa_string_get(IDSA_UNIT * u, void *p, int l)
{
  return _idsa_string_print(u, p, l, 0);
}

/* integer handler ****************************************************** */

static int _idsa_int_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  unsigned int x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(int));
  memcpy(&y, b->u_ptr, sizeof(int));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

static int _idsa_int_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(int));

  return 0;
}

static int _idsa_int_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(int)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(int));
    return sizeof(int);
  }
}

static int _idsa_int_scan(IDSA_UNIT * u, char *s)
{
  unsigned int i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(int));
    return 0;
  } else {
    return -1;
  }
}

static int _idsa_int_print(IDSA_UNIT * u, char *s, int l, int m)
{
  unsigned int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  return snprintf(s, l, "%u", i);
}

/* uid lookup: ********************************************************** */

static int _idsa_uid_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  uid_t x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(uid_t));
  memcpy(&y, b->u_ptr, sizeof(uid_t));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

static int _idsa_uid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(uid_t));

  return 0;
}

static int _idsa_uid_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(uid_t)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(uid_t));
    return sizeof(uid_t);
  }
}

static int _idsa_uid_scan(IDSA_UNIT * u, char *s)
{
  struct passwd *pw;
  uid_t i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(uid_t));
    return 0;
  } else {
    pw = getpwnam(s);
    if (pw) {
      memcpy(u->u_ptr, &(pw->pw_uid), sizeof(uid_t));
      return 0;
    } else {
      return 1;
    }
  }

  return 1;
}

static int _idsa_uid_print(IDSA_UNIT * u, char *s, int l, int m)
{
  uid_t i;

  memcpy(&i, u->u_ptr, sizeof(uid_t));

  return snprintf(s, l, "%u", i);
}

/* gid lookup: ********************************************************** */

static int _idsa_gid_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  gid_t x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(gid_t));
  memcpy(&y, b->u_ptr, sizeof(gid_t));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

static int _idsa_gid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(gid_t));

  return 0;
}

static int _idsa_gid_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(gid_t)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(gid_t));
    return sizeof(gid_t);
  }
}

static int _idsa_gid_scan(IDSA_UNIT * u, char *s)
{
  struct group *gr;
  gid_t i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(gid_t));
    return 0;
  } else {
    gr = getgrnam(s);
    if (gr) {
      memcpy(u->u_ptr, &(gr->gr_gid), sizeof(gid_t));
      return 0;
    } else {
      return 1;
    }
  }
}

static int _idsa_gid_print(IDSA_UNIT * u, char *s, int l, int m)
{
  gid_t i;

  memcpy(&i, u->u_ptr, sizeof(gid_t));

  return snprintf(s, l, "%u", i);
}

/* pid lookup: ********************************************************** */

static int _idsa_pid_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  pid_t x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(pid_t));
  memcpy(&y, b->u_ptr, sizeof(pid_t));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

static int _idsa_pid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(pid_t));

  return 0;
}

static int _idsa_pid_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(pid_t)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(pid_t));
    return sizeof(pid_t);
  }
}

static int _idsa_pid_scan(IDSA_UNIT * u, char *s)
{
  pid_t i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(pid_t));
    return 0;
  } else {
    return -1;
  }
}

static int _idsa_pid_print(IDSA_UNIT * u, char *s, int l, int m)
{
  pid_t i;

  memcpy(&i, u->u_ptr, sizeof(pid_t));

  return snprintf(s, l, "%u", i);
}

/* time handler (to be done properly) *********************************** */

static int _idsa_time_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  time_t x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(time_t));
  memcpy(&y, b->u_ptr, sizeof(time_t));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

static int _idsa_time_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(time_t));

  return 0;
}

static int _idsa_time_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(time_t)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(time_t));
    return sizeof(time_t);
  }
}

static int _idsa_time_scan(IDSA_UNIT * u, char *s)
{
  time_t i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(time_t));
    return 0;
  } else {
    return -1;
  }
}

static int _idsa_time_print(IDSA_UNIT * u, char *s, int l, int m)
{
  time_t i;

  memcpy(&i, u->u_ptr, sizeof(time_t));

  return snprintf(s, l, "%lu", i);
}

/* flag handler (nice and simple) *************************************** */

static int _idsa_flag_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  unsigned int x, y;

  memcpy(&x, a->u_ptr, sizeof(int));
  memcpy(&y, b->u_ptr, sizeof(int));

  if (x) {
    if (y) {
      return IDSA_COMPARE_EQUAL;
    } else {
      return IDSA_COMPARE_LESS;
    }
  } else {
    if (y) {
      return IDSA_COMPARE_MORE;
    } else {
      return IDSA_COMPARE_EQUAL;
    }
  }
}

static int _idsa_flag_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(int));

  return 0;
}

static int _idsa_flag_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < sizeof(unsigned int)) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, sizeof(unsigned int));
    return sizeof(unsigned int);
  }
}

static int _idsa_flag_scan(IDSA_UNIT * u, char *s)
{
  unsigned int i;

  if (s[0] == '0' || s[0] == 'f' || s[0] == 'F') {
    i = 0;
  } else {
    i = 1;
  }
  memcpy(u->u_ptr, &i, sizeof(int));

  return 0;
}

static int _idsa_flag_print(IDSA_UNIT * u, char *s, int l, int m)
{
  unsigned int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  if (l > 0) {
    if (i) {
      s[0] = '1';
    } else {
      s[0] = '0';
    }
    return 1;
  } else {
    return -1;
  }
}

/*
static int _idsa_flag_display(IDSA_UNIT * u, char *s, int l)
{
  unsigned int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  return snprintf(s, l, "%s", i ? "true" : "false");
}
*/

/* risk handler (not simple) ******************************************** */

static int _idsa_risk_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;

  unsigned short ab, ae, bb, be;

  memcpy(&ab, a->u_ptr, sizeof(short));
  memcpy(&ae, a->u_ptr + sizeof(short), sizeof(short));

  memcpy(&bb, b->u_ptr, sizeof(short));
  memcpy(&be, b->u_ptr + sizeof(short), sizeof(short));

  if (ae < bb) {
    result = IDSA_COMPARE_LESS;
  } else if (ab > be) {
    result = IDSA_COMPARE_MORE;
  } else if ((ab == bb) && (ae == be)) {
    result = IDSA_COMPARE_EQUAL;
  } else {
    result = IDSA_COMPARE_INTERSECT;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_compare(): [%d,%d] ? [%d,%d] yields %d\n",
	  ab, ae, bb, be, result);
#endif

  return result;
}

/*
static int _idsa_risk_check(IDSA_UNIT * u)
{
  return 0;
}
*/

static int _idsa_risk_set(IDSA_UNIT * u, void *p)
{
  double d;
  unsigned short v[2];

  d = *((double *) p);

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_set(): got %.3f, change to %.3f\n", d,
	  (d + 1.0) / 2.0);
#endif
  d = (d + 1.0) / 2.0;

  if ((d < -0.00001) || (d > 1.00001))
    return 1;

  v[0] = (unsigned short) (d * IDSA_M_RISKTOP);
  v[1] = (unsigned short) (d * IDSA_M_RISKTOP);

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_set(): [%04x-%04x]\n", v[0], v[1]);
#endif

  memcpy(u->u_ptr, v, 2 * sizeof(short));

  return 0;
}

static int _idsa_risk_get(IDSA_UNIT * u, void *p, int l)
{
  double d;
  unsigned short v[2];

  if (l < sizeof(double)) {
    return -1;
  } else {
    memcpy(v, u->u_ptr, sizeof(short) * 2);

    d = ((double) v[0]) / IDSA_M_RISKTOP;

    d = (d * 2.0) - 1.0;

#ifdef TRACE
    fprintf(stderr,
	    "_idsa_risk_get(): converted risk from <%hd> to <%f>\n", v[0],
	    d);
#endif

    memcpy(p, &d, sizeof(double));

    return sizeof(double);
  }
}

static int _idsa_risk_scan(IDSA_UNIT * u, char *s)
{
  int i;
  double a, b;
  unsigned short v[2];

  i = 0;

  if (isdigit(s[i]) || (s[i] == '-')) {
    a = (1.0 + atof(s + i)) / 2.0;
    b = a;
  } else {
    if (s[i] == '\0')
      return 1;

    i++;
    a = (1.0 + atof(s + i)) / 2.0;
    while ((s[i] != '\0') && (s[i] != '/') && (s[i] != ','))
      i++;
    if (s[i] == '\0')
      return 1;

    i++;
    b = (1.0 + atof(s + i)) / 2.0;
    while ((s[i] != '\0') && (s[i] != '>') && (s[i] != ']'))
      i++;
    if (s[i] == '\0')
      return 1;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_read(): floating %.2f-%.2f\n", a, b);
#endif

  if ((a < -0.00001) || (a > 1.00001) || (b < -0.00001) || (b > 1.00001))
    return 1;

  if (a > b) {
    v[0] = (unsigned short) (b * IDSA_M_RISKTOP);
    if (s[i] == '>')
      v[0] += 1;
    v[1] = (unsigned short) (a * IDSA_M_RISKTOP);
    if (s[0] == '<')
      v[1] -= 1;
  } else {
    v[0] = (unsigned short) (a * IDSA_M_RISKTOP);
    if (s[0] == '<')
      v[0] += 1;
    v[1] = (unsigned short) (b * IDSA_M_RISKTOP);
    if (s[i] == '>')
      v[1] -= 1;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_read(): [%04x-%04x]\n", v[0], v[1]);
#endif

  memcpy(u->u_ptr, v, 2 * sizeof(short));

  return 0;
}

static int _idsa_risk_print(IDSA_UNIT * u, char *s, int l, int m)
{
  unsigned short r;
  memcpy(&r, u->u_ptr, sizeof(short));

  return snprintf(s, l, "%.3f",
		  ((2.0 * (double) r) / (double) (IDSA_M_RISKTOP)) - 1.0);
}

/* errno handler ******************************************************** */

/*
static int _idsa_errno_display(IDSA_UNIT * u, char *s, int l)
{
  int i, n;
  char *p;

  memcpy(&i, u->u_ptr, sizeof(int));
  p = strerror(i);

#ifdef TRACE
  fprintf(stderr, "_idsa_errno_display(): errno is <%d>, string <%s>\n", i,
	  p);
#endif

  n = strlen(p);

  if (n > l) {
    memcpy(s, p, l);
    return -1;
  } else {
    memcpy(s, p, n);
    return n;
  }
}
*/

/* hostname handler (host.some.domain), stored in reverse order to ****** */
/* make matching of .domains easier ************************************* */

static int _idsa_host_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  int i, z;

  i = 0;
  z = idsa_type_size(a->u_type);
  while ((i < z) && (a->u_ptr[i] == b->u_ptr[i]) && (a->u_ptr[i] != '\0')) {
    i++;
  }

  if (a->u_ptr[i] == b->u_ptr[i]) {
    result = IDSA_COMPARE_EQUAL;
  } else {
    if (a->u_ptr[i] < b->u_ptr[i]) {
      result = IDSA_COMPARE_MORE;
    } else {			/* a > b */
      result = IDSA_COMPARE_LESS;
    }
    if (i > 0) {		/* check for trailing . for network match */
      if ((a->u_ptr[i] == '\0') || (b->u_ptr[i] == '\0')) {
	if ((b->u_ptr[i - 1] == '.')
	    && (a->u_ptr[i - 1] == b->u_ptr[i - 1])) {
	  result = IDSA_COMPARE_INTERSECT;
	}
      }
    }
  }

#ifdef TRACE
  fprintf(stderr,
	  "compare_host(): compare of <%s> against <%s> yields <%d>\n",
	  a->u_ptr, b->u_ptr, result);
#endif

  return result;
}

static int _idsa_host_check(IDSA_UNIT * u)
{
  u->u_ptr[idsa_type_size(u->u_type) - 1] = '\0';
  return 0;
}

static int _idsa_host_scan(IDSA_UNIT * u, char *s)
{
  int i, j, z;
  int result = 0;

  i = strlen(s);
  z = idsa_type_size(u->u_type);
  if (i > (z - 1)) {
    i = z - 1;
    result = 1;
  }
  u->u_ptr[i] = '\0';
  i--;

  j = 0;
  while (i >= 0) {
    u->u_ptr[j] = s[i];
    j++;
    i--;
  }

  return result;
}

static int _idsa_host_print(IDSA_UNIT * u, char *s, int l, int m)
{
  int i, j, result, z;

  i = 0;
  z = idsa_type_size(u->u_type);
  while ((i < z) && (u->u_ptr[i] != '\0')) {
    i++;
  }

  if (l < i) {
    result = (-1);
    i = l;
  } else {
    result = i;
  }
#ifdef TRACE
  fprintf(stderr, "print_host(): <%s:%d>\n", u->u_ptr, i);
#endif
  i--;

  j = 0;
  while (i >= 0) {
    s[j] = u->u_ptr[i];
    j++;
    i--;
  }

  return result;
}

static int _idsa_host_set(IDSA_UNIT * u, void *p)
{
  return _idsa_host_scan(u, p);
}

static int _idsa_host_get(IDSA_UNIT * u, void *p, int l)
{
  return _idsa_host_print(u, p, l, 0);
}

/* inet ipv4 address handler. Internal storage representation is an ***** */
/* array of 2 integers, one for address, and one for (32-mask) ********** */

static int _idsa_ip4addr_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  unsigned long int p[2], q[2], m;

  memcpy(p, a->u_ptr, 2 * sizeof(long int));
  memcpy(q, b->u_ptr, 2 * sizeof(long int));
  if (p[1] > q[1]) {
    m = (0xffffffff << (p[1]));
  } else {
    m = (0xffffffff << (q[1]));
  }

  p[0] = (p[0] & m);
  q[0] = (q[0] & m);

  if (p[0] > q[0]) {
    result = IDSA_COMPARE_LESS;
  } else if (p[0] < q[0]) {
    result = IDSA_COMPARE_LESS;
  } else {
    if (p[1] == q[1]) {
      result = IDSA_COMPARE_EQUAL;
    } else {
      result = IDSA_COMPARE_INTERSECT;
    }
  }

#ifdef TRACE
  fprintf(stderr,
	  "_idsa_ip4addr_compare(): hostmask is <0x%08lx> bits, <0x%08lx?0x%08lx>, result <%d>\n",
	  m, p[0], q[0], result);
#endif

  return result;
}

static int _idsa_ip4addr_check(IDSA_UNIT * u)
{
  unsigned long int a[2];
  int result = 0;

  memcpy(a, u->u_ptr, 2 * sizeof(long int));
  if (a[1] > 32) {
    a[1] = 0;
    result++;
  }
  return result;
}

static int _idsa_ip4addr_set(IDSA_UNIT * u, void *p)
{
  unsigned long int a[2];

  memcpy(a, p, sizeof(long int));
  a[1] = 0;

#ifdef TRACE
  fprintf(stderr, "_idsa_ip4addr_set(): 0x%08lx/%ld\n", a[0], a[1]);
#endif

  memcpy(u->u_ptr, a, 2 * sizeof(long int));

  return 0;
}

static int _idsa_ip4addr_get(IDSA_UNIT * u, void *p, int l)
{
  unsigned long int a[2];

  memcpy(a, u->u_ptr, 2 * sizeof(long int));

#ifdef TRACE
  fprintf(stderr, "_idsa_ip4addr_get(): 0x%08lx/%ld\n", a[0], a[1]);
#endif

  if (l < sizeof(long int)) {
    return -1;
  } else {
    memcpy(p, a, sizeof(long int));
    return sizeof(long int);
  }
}

static int _idsa_ip4addr_scan(IDSA_UNIT * u, char *s)
{
  struct in_addr d;
  int result;
  int i;
  unsigned long int a[2];
  char tmp[16];

  a[0] = 0;
  a[1] = 0;

  i = 0;
  while ((s[i] != '\0') && (s[i] != '/')) {
    i++;
  }

  result = 0;
  if (s[i] == '/') {
    if (i > 15) {
      i = 15;
    }
    memcpy(tmp, s, i);
    tmp[i] = '\0';
    if (inet_aton(tmp, &d) == 0) {
      result++;
    }
    a[1] = 32 - atoi(s + i + 1);
#ifdef TRACE
    fprintf(stderr,
	    "_idsa_ip4addr_scan(): address component is <%s>, number %ld\n",
	    tmp, a[1]);
#endif
    if (a[1] > 32) {
      result++;
    }
  } else {			/* no /bits component, just plain IP address */
    if (inet_aton(s, &d) == 0) {
      result++;
    }
  }
  a[0] = ntohl(d.s_addr);

#ifdef TRACE
  fprintf(stderr, "_idsa_ip4addr_scan(): 0x%08lx/%ld\n", a[0], a[1]);
#endif

  memcpy(u->u_ptr, a, 2 * sizeof(long int));

  return result;
}

static int _idsa_ip4addr_print(IDSA_UNIT * u, char *s, int l, int m)
{
  unsigned long int a[2];
  struct in_addr d;
  int result;

  memcpy(a, u->u_ptr, 2 * sizeof(long int));
  d.s_addr = htonl(a[0]);

  if (a[1] > 0) {
    return snprintf(s, l, "%s/%ld", inet_ntoa(d), 32 - a[1]);
  } else {
    return snprintf(s, l, "%s", inet_ntoa(d));
  }

  return result;
}

/* port handler (protocol/port) ***************************************** */

static int _idsa_ipport_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int x[2], y[2];
  int result;

  memcpy(&x, a->u_ptr, 2 * sizeof(int));
  memcpy(&y, b->u_ptr, 2 * sizeof(int));

  if (x[0] < y[0]) {
    result = IDSA_COMPARE_LESS;
  } else if (x[0] > y[0]) {
    result = IDSA_COMPARE_MORE;
  } else {
    if (x[1] < y[1]) {
      result = IDSA_COMPARE_LESS;
    } else if (x[1] > y[1]) {
      result = IDSA_COMPARE_MORE;
    } else {
      result = IDSA_COMPARE_EQUAL;
    }
  }

  return result;
}

static int _idsa_ipport_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, 2 * sizeof(int));

  return 0;
}

static int _idsa_ipport_get(IDSA_UNIT * u, void *p, int l)
{
  if (l < (2 * sizeof(int))) {
    return -1;
  } else {
    memcpy(p, u->u_ptr, 2 * sizeof(int));
    return (2 * sizeof(int));
  }
}

static int _idsa_ipport_scan(IDSA_UNIT * u, char *s)
{
  int x[2];
  int result = 0;
  char p[32];
  int i;
  struct protoent *pe;
  struct servent *se;

  for (i = 0; (s[i] != '\0') && (s[i] != '/'); i++);

  if ((i < 31) && (s[i] == '/')) {
    memcpy(p, s, i);
    p[i] = '\0';
    i++;

    if (isdigit(p[0])) {
      x[0] = atoi(p);
    } else {
      pe = getprotobyname(p);
      if (pe) {
	x[0] = pe->p_proto;
      } else {
	x[0] = 0;
	result++;
      }
    }

    if (isdigit(s[i])) {
      x[1] = atoi(s + i);
    } else {
      se = getservbyname(s + i, p);
      if (se) {
	x[1] = ntohs(se->s_port);
      } else {
	x[1] = 0;
	result++;
      }
    }

#ifdef TRACE
    fprintf(stderr, "_idsa_ipport_scan(): <%s> maps to %d/%d\n", s, x[0],
	    x[1]);
#endif
    memcpy(u->u_ptr, x, 2 * sizeof(int));

  } else {
    result++;
  }

  return result;
}

static int _idsa_ipport_print(IDSA_UNIT * u, char *s, int l, int m)
{
  int x[2];
  memcpy(&x, u->u_ptr, 2 * sizeof(int));

  return snprintf(s, l, "%d/%d", x[0], x[1]);
}

/* port handler (protocol/port) ***************************************** */

static int _idsa_file_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  int i, z;

  i = 0;
  z = idsa_type_size(a->u_type);
  while ((i < z) && (a->u_ptr[i] == b->u_ptr[i]) && (a->u_ptr[i] != '\0')) {
    i++;
  }

  if ((a->u_ptr[i] == '\0') || (b->u_ptr[i] == '\0')) {
    if (a->u_ptr[i] == b->u_ptr[i]) {
      result = IDSA_COMPARE_EQUAL;
    } else {
      if (a->u_ptr[i] == '\0') {
	if ((i > 0) && (a->u_ptr[i - 1] == '/')) {
	  result = IDSA_COMPARE_INTERSECT;
	} else {
	  result = IDSA_COMPARE_LESS;
	}
      } else {
	if ((i > 0) && (b->u_ptr[i - 1] == '/')) {
	  result = IDSA_COMPARE_INTERSECT;
	} else {
	  result = IDSA_COMPARE_MORE;
	}
      }
    }
  } else {
    if (a->u_ptr[i] < b->u_ptr[i]) {
      result = IDSA_COMPARE_LESS;
    } else {
      result = IDSA_COMPARE_MORE;
    }
  }

#ifdef TRACE
  fprintf(stderr, "compare_file(): comparing of <%s> against <%s>\n",
	  a->u_ptr, b->u_ptr);
#endif

  return result;
}

/****************************************************************************/

struct idsa_type_details {
  unsigned int l_type;
  unsigned int l_size;
  unsigned int l_cform;
  char l_name[IDSA_M_NAME];
  IDSA_FUNCTION_COMPARE l_compare;
  IDSA_FUNCTION_CHECK l_check;
  IDSA_FUNCTION_SET l_set;
  IDSA_FUNCTION_GET l_get;
  IDSA_FUNCTION_SCAN l_scan;
  IDSA_FUNCTION_PRINT l_print;
};

static IDSA_TYPE_DETAILS _idsa_type_table[IDSA_M_TYPES] = {
  [0] = {IDSA_T_NULL,
	 0,
	 0,
	 "",
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 NULL},

  [IDSA_T_STRING] = {IDSA_T_STRING,
		     IDSA_M_MEDIUM,
		     0,
		     "string",
		     &_idsa_string_compare,
		     &_idsa_string_check,
		     &_idsa_string_set,
		     &_idsa_string_get,
		     &_idsa_string_scan,
		     &_idsa_string_print},

  [IDSA_T_INT] = {IDSA_T_INT,
		  sizeof(int),
		  0,
		  "integer",
		  &_idsa_int_compare,
		  NULL,
		  &_idsa_int_set,
		  &_idsa_int_get,
		  &_idsa_int_scan,
		  &_idsa_int_print},

  [IDSA_T_UID] = {IDSA_T_UID,
		  sizeof(uid_t),
		  0,
		  "uid",
		  &_idsa_uid_compare,
		  NULL,
		  &_idsa_uid_set,
		  &_idsa_uid_get,
		  &_idsa_uid_scan,
		  &_idsa_uid_print}
  ,

  [IDSA_T_GID] = {IDSA_T_GID,
		  sizeof(gid_t),
		  0,
		  "gid",
		  &_idsa_gid_compare,
		  NULL,
		  &_idsa_gid_set,
		  &_idsa_gid_get,
		  &_idsa_gid_scan,
		  &_idsa_gid_print}
  ,

  [IDSA_T_PID] = {IDSA_T_PID,
		  sizeof(pid_t),
		  0,
		  "pid",
		  &_idsa_pid_compare,
		  NULL,
		  &_idsa_pid_set,
		  &_idsa_pid_get,
		  &_idsa_pid_scan,
		  &_idsa_pid_print}
  ,

  [IDSA_T_TIME] = {IDSA_T_TIME,
		   sizeof(time_t),
		   0,
		   "time",
		   &_idsa_time_compare,
		   NULL,
		   &_idsa_time_set,
		   &_idsa_time_get,
		   &_idsa_time_scan,	/* needs improvement */
		   &_idsa_time_print}
  ,

  [IDSA_T_FLAG] = {IDSA_T_FLAG,
		   sizeof(int),
		   0,
		   "flag",
		   &_idsa_flag_compare,
		   NULL,
		   &_idsa_flag_set,
		   &_idsa_flag_get,
		   &_idsa_flag_scan,
		   &_idsa_flag_print},

  [IDSA_T_RISK] = {IDSA_T_RISK,
		   2 * sizeof(short),
		   0,
		   "risk",
		   &_idsa_risk_compare,
		   NULL,
		   &_idsa_risk_set,
		   &_idsa_risk_get,
		   &_idsa_risk_scan,
		   &_idsa_risk_print},

  [IDSA_T_ERRNO] = {IDSA_T_ERRNO,
		    sizeof(int),
		    0,
		    "errno",
		    &_idsa_int_compare,
		    NULL,
		    &_idsa_int_set,
		    &_idsa_int_get,
		    &_idsa_int_scan,
		    &_idsa_int_print},

  [IDSA_T_HOST] = {IDSA_T_HOST,
		   IDSA_M_MEDIUM,
		   0,
		   "host",
		   &_idsa_host_compare,
		   &_idsa_host_check,
		   &_idsa_host_set,
		   &_idsa_host_get,
		   &_idsa_host_scan,
		   &_idsa_host_print},

  [IDSA_T_ADDR] = {IDSA_T_ADDR,
		   2 * sizeof(unsigned long),
		   0,
		   "addr",
		   &_idsa_ip4addr_compare,
		   &_idsa_ip4addr_check,
		   &_idsa_ip4addr_set,
		   &_idsa_ip4addr_get,
		   &_idsa_ip4addr_scan,
		   &_idsa_ip4addr_print},

  [IDSA_T_PORT] = {IDSA_T_PORT,
		   2 * sizeof(int),
		   0,
		   "port",
		   &_idsa_ipport_compare,
		   NULL,
		   &_idsa_ipport_set,
		   &_idsa_ipport_get,
		   &_idsa_ipport_scan,
		   &_idsa_ipport_print},

  [IDSA_T_FILE] = {IDSA_T_FILE,
		   IDSA_M_MEDIUM,
		   0,
		   "file",
		   &_idsa_file_compare,
		   &_idsa_string_check,
		   &_idsa_string_set,
		   &_idsa_string_get,
		   &_idsa_string_scan,
		   &_idsa_string_print}
};

int idsa_unit_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  IDSA_TYPE_DETAILS *l;

  l = idsa_type_lookup(a->u_type);
  if (l) {
    return (*(l->l_compare)) (a, b);
  } else {
    return IDSA_COMPARE_LESS;
  }
}

int idsa_unit_check(IDSA_UNIT * u)
{
  IDSA_TYPE_DETAILS *t;
  int i;

  u->u_name[IDSA_M_NAME - 1] = '\0';
  for (i = 0; u->u_name[i] != '\0'; i++) {
    if (!isgraph(u->u_name[i])) {
      u->u_name[i] = '_';
    }
  }

  t = idsa_type_lookup(u->u_type);
  if (t) {
    if (t->l_check) {
      return (*(t->l_check)) (u);
    } else {
      return 0;
    }
  } else {
    return 1;
  }
}

int idsa_unit_get(IDSA_UNIT * u, void *p, int l)
{
  IDSA_TYPE_DETAILS *t;

#ifdef TRACE
  fprintf(stderr, "idsa_unit_get(): looking up type <0x%04x>\n",
	  u->u_type);
#endif

  t = idsa_type_lookup(u->u_type);
  if (t) {
    return (*(t->l_get)) (u, p, l);
  } else {
    return 1;
  }
}

int idsa_unit_set(IDSA_UNIT * u, void *p)
{
  IDSA_TYPE_DETAILS *t;

#ifdef TRACE
  fprintf(stderr, "idsa_unit_set(): setting type <0x%04x> (address %p)\n",
	  u->u_type, p);
#endif

  t = idsa_type_lookup(u->u_type);
  if (t) {
    return (*(t->l_set)) (u, p);
  } else {
    return 1;
  }
}

int idsa_unit_scan(IDSA_UNIT * u, char *s)
{
  IDSA_TYPE_DETAILS *t;

  t = idsa_type_lookup(u->u_type);
  if (t) {
    return (*(t->l_scan)) (u, s);
  } else {
    return 1;
  }
}

int idsa_unit_print(IDSA_UNIT * u, char *s, int l, int m)
{
  IDSA_TYPE_DETAILS *t;

  t = idsa_type_lookup(u->u_type);
  if (t) {
    return (*(t->l_print)) (u, s, l, m);
  } else {
    return -1;
  }
}

/****************************************************************************/

IDSA_TYPE_DETAILS *idsa_type_lookup(unsigned int t)
{
  IDSA_TYPE_DETAILS *result;
  unsigned int i;

  i = t;
  if (i > 0 && i < IDSA_M_TYPES) {
    result = &(_idsa_type_table[i]);
    if (result->l_type != t) {
#ifdef TRACE
      fprintf(stderr,
	      "idsa_type_lookup(): sanity failure, type table corruption - have <0x%04x>, want <0x%04x>\n",
	      result->l_type, t);
#endif
      result = NULL;
    }
  } else {
#ifdef TRACE
    fprintf(stderr, "idsa_type_lookup(): unknown type 0x%04x\n", t);
#endif
    result = NULL;

  }

  return result;
}

char *idsa_type_name(unsigned int t)
{
  IDSA_TYPE_DETAILS *l;

  l = idsa_type_lookup(t);
  if (l) {
    return l->l_name;
  } else {
    return NULL;
  }
}

unsigned int idsa_type_code(char *n)
{
  int i;

  for (i = 1; i < IDSA_M_TYPES; i++) {
#ifdef TRACE
    fprintf(stderr, "idsa_type_code(): looking at <%s:0x%04x>\n",
	    _idsa_type_table[i].l_name, _idsa_type_table[i].l_type);
#endif
    if (strcmp(n, _idsa_type_table[i].l_name) == 0) {
      return _idsa_type_table[i].l_type;
    }
  }
  return IDSA_T_NULL;
}

int idsa_type_size(unsigned int t)
{
  IDSA_TYPE_DETAILS *l;

  l = idsa_type_lookup(t);
  if (l) {
    return l->l_size;
  } else {
    return 0;
  }
}
