#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <idsa_internal.h>

#include "mex.h"

#include "set.h"
#include "rule.h"
#include "job.h"
#include "table.h"
#include "parse.h"
#include "output.h"
#include "queue.h"

static void match_dump(RULE_MATCH * m, FILE * f, int c);
static void match_fire_up(RULE_MATCH * m);

void rule_clear(RULE_BODY * r)
{
  RULE_MATCH *m, *n;

  while (r) {
    m = r->r_match;
    n = NULL;

    while (m != NULL) {
#ifdef TRACE
/* fprintf(stderr, "rule_clear(): looking at %p, previous %p\n", m, n); */
#endif
      if (m->m_left && (m->m_up == n)) {	/* come from top, move left */
	n = m;
	m = m->m_left;
      } else if (m->m_right && (m->m_right == n)) {	/* come from right, move up */
	n = m;
	m = m->m_up;
      } else {
	m->m_fire = RULE_FIRE_CLEAR;
	n = m;
	if (m->m_right) {
	  m = m->m_right;
	} else {
	  m = m->m_up;
	}
      }
    }
    r = r->r_next;
  }
}

int rule_run(STATE_SET * s, RULE_BODY * r, IDSA_EVENT * em,
	     IDSA_EVENT * er, JOB * j)
{
  RULE_ACTION *a;

#ifdef TRACE
  fprintf(stderr, "rule_run(): looking at rule %p\n", r);
#endif

  a = r->r_action;
  while (a) {
    switch (a->a_type) {
    case RULE_ACTION_LOG:
#ifdef TRACE
      fprintf(stderr, "rule_run(): should log event\n");
#endif
      s->s_outputfail += output_run(a->a_un.a_output, em);
      break;
    case RULE_ACTION_SEND:
      if (er) {
	idsa_add_unit(er, a->a_un.a_unit);
      }
      break;
    case RULE_ACTION_KEEP:
      queue_run(a->a_un.a_keep, em);
      break;
    default:			/* major runtime failure which should not happen */
#ifdef TRACE
      fprintf(stderr, "rule_run(): unknown action type 0x%02x\n",
	      a->a_type);
#endif
      s->s_outputfail = 1;
      break;
    }

    a = a->a_next;
  }

  return r->r_policy;
}

RULE_BODY *rule_new()
{
  RULE_BODY *r;

  r = malloc(sizeof(RULE_BODY));
  if (r) {
    r->r_policy = RULE_ACTION_NULL;
    r->r_match = NULL;
    r->r_action = NULL;
    r->r_next = NULL;
  }
  return r;
}

void rule_free(RULE_BODY * r)
{
  RULE_ACTION *a, *b;

#ifdef TRACE
  fprintf(stderr, "rule_free(): deleting rule %p\n", r);
#endif

  if (r->r_match) {
    match_free(r->r_match);
    r->r_match = NULL;
  }
  a = r->r_action;
  while (a) {
    b = a;
    a = a->a_next;
#ifdef TRACE
    fprintf(stderr, "rule_free(): deleting action %p\n", b);
#endif
    switch (b->a_type) {
    case RULE_ACTION_LOG:
      /* logs get deallocated somewhere else */
      break;
    case RULE_ACTION_SEND:
      idsa_unit_free(b->a_un.a_unit);
      break;
    case RULE_ACTION_KEEP:
      queue_keep_free(b->a_un.a_keep);
      break;
    }
    free(b);
  }

  free(r);
}

RULE_MATCH *match_new(int g)
{
  RULE_MATCH *m;

#ifdef TRACE
  fprintf(stderr, "match_new(): allocating <%d> bytes\n",
	  sizeof(RULE_MATCH));
#endif

  m = malloc(sizeof(RULE_MATCH));
  if (m) {
    m->m_gate = g;
    m->m_fire = RULE_FIRE_CLEAR;
    m->m_up = NULL;
    m->m_left = NULL;
    m->m_right = NULL;
  }
  return m;
}

int match_fire_true(RULE_MATCH * m)
{

  if (m->m_gate & RULE_GATE_NEGATE) {
    m->m_fire = RULE_FIRE_FALSE;
  } else {
    m->m_fire = RULE_FIRE_TRUE;
  }

  match_fire_up(m);

  return 0;
}

int match_fire_false(RULE_MATCH * m)
{
  if (m->m_gate & RULE_GATE_NEGATE) {
    m->m_fire = RULE_FIRE_TRUE;
  } else {
    m->m_fire = RULE_FIRE_FALSE;
  }

  match_fire_up(m);

  return 0;
}

void match_free(RULE_MATCH * m)
{
  RULE_MATCH *n;

  /* postorder deletion */
  while (m) {
    if (m->m_left != NULL) {
      m = m->m_left;
    } else if (m->m_right != NULL) {
      m = m->m_right;
    } else {
      n = m;
      m = m->m_up;
      if (m) {
	if (m->m_left == n) {
	  m->m_left = NULL;
	} else {
	  m->m_right = NULL;
	}
      }
      n->m_up = NULL;
      n->m_left = NULL;
      n->m_right = NULL;
#ifdef TRACE
      fprintf(stderr, "match_free(): deleting match %p\n", n);
#endif
      free(n);
    }
  }
}

RULE_ACTION *action_new(int t)
{
  RULE_ACTION *a;

  a = malloc(sizeof(RULE_ACTION));
  if (a) {
    a->a_type = t;
    switch (a->a_type) {
    case RULE_ACTION_LOG:
      a->a_un.a_output = NULL;
      break;
    case RULE_ACTION_SEND:
      a->a_un.a_unit = NULL;
      break;
    case RULE_ACTION_KEEP:
      a->a_un.a_keep = NULL;
      break;
    }
    a->a_next = NULL;
  }
  return a;
}

void action_free(RULE_ACTION * a)
{
  a->a_next = NULL;
  free(a);
}

static void match_fire_up(RULE_MATCH * m)
{
  RULE_MATCH *n, *o;

  n = m->m_up;

  while (n && (n->m_fire == RULE_FIRE_CLEAR)) {
#ifdef TRACE
    fprintf(stderr,
	    "match_fire_up(): firing up, me[%p]=%s, parent[%p]=%d, left[%p]=%d, right[%p]=%d, up[%p]\n",
	    m, (m->m_fire == RULE_FIRE_TRUE) ? "true" : "false", n,
	    n->m_gate, n->m_left, n->m_left->m_fire, n->m_right,
	    n->m_right->m_fire, n->m_up);
#endif
    switch (n->m_gate) {
    case RULE_GATE_AND:
      if (m->m_fire == RULE_FIRE_TRUE) {
	o = (n->m_left == m) ? (n->m_right) : (n->m_left);
	switch (o->m_fire) {
	case RULE_FIRE_CLEAR:
	  n = NULL;
	  break;
	case RULE_FIRE_TRUE:
	  n->m_fire = RULE_FIRE_TRUE;
	  m = n;
	  n = n->m_up;
	  break;
	case RULE_FIRE_FALSE:
	  n->m_fire = RULE_FIRE_FALSE;
	  m = n;
	  n = n->m_up;
	  break;
	}
      } else {
	n->m_fire = RULE_FIRE_FALSE;
	m = n;
	n = n->m_up;
      }
      break;
    case RULE_GATE_OR:
      if (m->m_fire == RULE_FIRE_TRUE) {
	n->m_fire = RULE_FIRE_TRUE;
	m = n;
	n = n->m_up;
      } else {
	o = (n->m_left == m) ? (n->m_right) : (n->m_left);
	switch (o->m_fire) {
	case RULE_FIRE_CLEAR:
	  n = NULL;
	  break;
	case RULE_FIRE_TRUE:
	  n->m_fire = RULE_FIRE_TRUE;
	  m = n;
	  n = n->m_up;
	  break;
	case RULE_FIRE_FALSE:
	  n->m_fire = RULE_FIRE_FALSE;
	  m = n;
	  n = n->m_up;
	  break;
	}
      }
      break;
    case RULE_GATE_NAND:
      if (m->m_fire == RULE_FIRE_TRUE) {
	o = (n->m_left == m) ? (n->m_right) : (n->m_left);
	switch (o->m_fire) {
	case RULE_FIRE_CLEAR:
	  n = NULL;
	  break;
	case RULE_FIRE_TRUE:
	  n->m_fire = RULE_FIRE_FALSE;
	  m = n;
	  n = n->m_up;
	  break;
	case RULE_FIRE_FALSE:
	  n->m_fire = RULE_FIRE_TRUE;
	  m = n;
	  n = n->m_up;
	  break;
	}
      } else {
	n->m_fire = RULE_FIRE_TRUE;
	m = n;
	n = n->m_up;
      }
      break;
    case RULE_GATE_NOR:
      if (m->m_fire == RULE_FIRE_TRUE) {
	n->m_fire = RULE_FIRE_FALSE;
	m = n;
	n = n->m_up;
      } else {
	o = (n->m_left == m) ? (n->m_right) : (n->m_left);
	switch (o->m_fire) {
	case RULE_FIRE_CLEAR:
	  n = NULL;
	  break;
	case RULE_FIRE_TRUE:
	  n->m_fire = RULE_FIRE_FALSE;
	  m = n;
	  n = n->m_up;
	  break;
	case RULE_FIRE_FALSE:
	  n->m_fire = RULE_FIRE_TRUE;
	  m = n;
	  n = n->m_up;
	  break;
	}
      }
      break;
    default:			/* should never happend, but just to prevent loops */
      n = NULL;
      break;
    }
  }
}

static void match_dump(RULE_MATCH * m, FILE * f, int c)
{
  int i;
  for (i = 0; i < c; i++) {
    fputc(' ', f);
  }

  fprintf(f, "%d (", m->m_fire);

  fputc('(', f);
  if (m->m_left) {
    fprintf(f, "left=%p ", m->m_left);
  }
  if (m->m_right) {
    fprintf(f, "right=%p", m->m_right);
  }
  fputc(')', f);

  if (m->m_gate & RULE_GATE_NEGATE) {
    fprintf(f, " !");
  }
  if (m->m_gate & RULE_GATE_AND) {
    fprintf(f, " &\n");
  } else if (m->m_gate & RULE_GATE_OR) {
    fprintf(f, " |\n");
  } else {
    fprintf(f, " [%p]\n", m);
  }
  if (m->m_left) {
    match_dump(m->m_left, f, c + 2);
  }
  if (m->m_right) {
    match_dump(m->m_right, f, c + 2);
  }
}

void rule_dump(RULE_BODY * r, FILE * f)
{
  RULE_ACTION *a;

  fprintf(f, "Rule %d, policy: 0x%04x\n", r->r_number, r->r_policy);

  fprintf(f, "Match:\n");
  if (r->r_match) {
    match_dump(r->r_match, f, 0);
  }
  a = r->r_action;
  while (a) {
    fprintf(f, "Action: type 0x%04x\n", a->a_type);
    a = a->a_next;
  }

}
